@@ -1,6 +1,7 @@
 import time
 import six
 
+from . import exceptions
 from . import packet
 from . import payload
 
@@ -29,7 +30,7 @@ class Socket(object):
             packets = [self.queue.get(timeout=self.server.ping_timeout)]
             self.queue.task_done()
         except self.server._async['queue'].Empty:
-            raise IOError()
+            raise exceptions.QueueEmpty()
         if packets == [None]:
             return []
         try:
@@ -58,7 +59,7 @@ class Socket(object):
         elif pkt.packet_type == packet.CLOSE:
             self.close(wait=False, abort=True)
         else:
-            raise ValueError
+            raise exceptions.UnknownPacketError()
 
     def send(self, pkt):
         """Send a packet to the client."""
@@ -88,7 +89,7 @@ class Socket(object):
                                                           start_response)
         try:
             packets = self.poll()
-        except IOError as e:
+        except exceptions.QueueEmpty as e:
             self.close(wait=False)
             raise e
         return packets
@@ -97,7 +98,7 @@ class Socket(object):
         """Handle a long-polling POST request from the client."""
         length = int(environ.get('CONTENT_LENGTH', '0'))
         if length > self.server.max_http_buffer_size:
-            raise ValueError()
+            raise exceptions.ContentTooLongError()
         else:
             body = environ['wsgi.input'].read(length)
             p = payload.Payload(encoded_payload=body)
@@ -108,12 +109,18 @@ class Socket(object):
         """Close the socket connection."""
         if not self.closed and not self.closing:
             self.closing = True
-            self.server._trigger_event('disconnect', self.sid, async=False)
+            reraise_exc = None
+            try:
+                self.server._trigger_event('disconnect', self.sid, async=False)
+            except Exception as e:
+                reraise_exc = e
             if not abort:
                 self.send(packet.Packet(packet.CLOSE))
             self.closed = True
             if wait:
                 self.queue.join()
+            if reraise_exc:
+                raise reraise_exc
 
     def _upgrade_websocket(self, environ, start_response):
         """Upgrade the connection from polling to websocket."""
@@ -130,6 +137,11 @@ class Socket(object):
 
     def _websocket_handler(self, ws):
         """Engine.IO handler for websocket transport."""
+        # try to set a socket timeout matching the configured ping interval
+        for attr in ['_sock', 'socket']:  # pragma: no cover
+            if hasattr(ws, attr) and hasattr(ws.attr, 'settimeout'):
+                getattr(ws, attr).settimeout(self.server.ping_interval)
+
         if self.connected:
             # the socket was already connected, so this is an upgrade
             self.queue.join()  # flush the queue first
@@ -163,9 +175,10 @@ class Socket(object):
         # start separate writer thread
         def writer():
             while True:
+                packets = None
                 try:
                     packets = self.poll()
-                except IOError:
+                except exceptions.QueueEmpty:
                     break
                 if not packets:
                     # empty packet list returned -> connection closed
@@ -175,16 +188,23 @@ class Socket(object):
                         ws.send(pkt.encode(always_bytes=False))
                 except:
                     break
-        self.server.start_background_task(writer)
+        writer_task = self.server.start_background_task(writer)
 
         self.server.logger.info(
             '%s: Upgrade to websocket successful', self.sid)
 
+        reraise_exc = None
         while True:
             p = None
             try:
                 p = ws.wait()
-            except:
+            except Exception as e:
+                # if the socket is already closed, we can assume this is a
+                # downstream error of that
+                if not self.closed:  # pragma: no cover
+                    self.server.logger.info(
+                        '%s: Unexpected error "%s", closing connection',
+                        self.sid, str(e))
                 break
             if p is None:
                 # connection closed by client
@@ -194,10 +214,19 @@ class Socket(object):
             pkt = packet.Packet(encoded_packet=p)
             try:
                 self.receive(pkt)
-            except ValueError:
+            except exceptions.UnknownPacketError:
                 pass
+            except Exception as e:
+                # if we get an unexpected exception (such as something in an
+                # application event handler) we close the connection properly
+                # and then reraise the exception
+                reraise_exc = e
+                break
 
-        self.close(wait=True, abort=True)
         self.queue.put(None)  # unlock the writer task so that it can exit
+        writer_task.join()
+        self.close(wait=True, abort=True)
+        if reraise_exc:
+            raise reraise_exc
 
         return []
